"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.main = main;
const child_process_1 = require("child_process");
const fs_1 = require("fs");
const fs_2 = __importDefault(require("fs"));
const path_1 = require("path");
const semver_1 = __importDefault(require("semver"));
const os_1 = require("os");
function main(options) {
    let storeDir = (0, path_1.resolve)(options.storeDir);
    (0, fs_1.mkdirSync)(storeDir, { recursive: true });
    let storePackageVersions = scanStorePackages(storeDir);
    let collectedNodeModules = new Set();
    let linkedDepPackageDirs = new Set();
    let context = {
        options,
        storeDir,
        storePackageVersions,
        collectedNodeModules,
        linkedDepPackageDirs,
    };
    initPackageFile(options.cwd);
    if (options.recursive) {
        scanPackageRecursively(context, options.cwd, new Set());
        return;
    }
    installPackages(context, options.cwd);
}
function scanStorePackages(storeDir) {
    // package name -> exact versions
    let storePackageVersions = new Map();
    for (let dirname of (0, fs_1.readdirSync)(storeDir)) {
        if (dirname[0] !== '@') {
            let [name, version] = dirname.split('@');
            getVersions(storePackageVersions, name).add(version);
            continue;
        }
        let orgName = dirname;
        let orgDir = (0, path_1.join)(storeDir, dirname);
        for (let dirname of (0, fs_1.readdirSync)(orgDir)) {
            let [name, version] = dirname.split('@');
            name = `${orgName}/${name}`;
            getVersions(storePackageVersions, name).add(version);
        }
    }
    return storePackageVersions;
}
function initPackageFile(packageDir) {
    let packageFile = (0, path_1.join)(packageDir, 'package.json');
    if (!(0, fs_1.existsSync)(packageFile)) {
        (0, fs_1.writeFileSync)(packageFile, '{}');
    }
}
function installPackages(context, packageDir) {
    let { storeDir, storePackageVersions, options, collectedNodeModules, linkedDepPackageDirs, } = context;
    let packageFile = (0, path_1.join)(packageDir, 'package.json');
    let packageJson = JSON.parse((0, fs_1.readFileSync)(packageFile).toString());
    let { dependencies, devDependencies } = packageJson;
    let nodeModulesDir = (0, path_1.join)(packageDir, 'node_modules');
    (0, fs_1.mkdirSync)(nodeModulesDir, { recursive: true });
    let usedPackageVersions = new Map();
    let newDeps = {};
    let hasNewDeps = false;
    let newInstallDeps = {};
    function installDeps(deps, installDeps) {
        function registerVersion(name, version) {
            deps[name] = version;
            newInstallDeps[name] = version;
        }
        installDeps: for (let dep of installDeps) {
            if (isLinkDep(dep)) {
                let name = null;
                name = linkDepPackage(linkedDepPackageDirs, nodeModulesDir, name, dep);
                getVersions(usedPackageVersions, name).add(dep);
                deps[name] = dep;
                continue;
            }
            let { name, version } = parseDep(dep);
            let storeVersions = getVersions(storePackageVersions, name);
            let exactVersion = semver_1.default.maxSatisfying(Array.from(storeVersions), version || '*');
            if (exactVersion) {
                linkStorePackage(storeDir, nodeModulesDir, name, exactVersion);
                registerVersion(name, `${exactVersion}`);
                continue;
            }
            let npmVersions = npmViewVersions(dep);
            if (npmVersions.length === 0)
                throw new Error('No versions found: ' + dep);
            npmVersions.reverse();
            for (let exactVersion of npmVersions) {
                if (storeVersions.has(exactVersion)) {
                    linkStorePackage(storeDir, nodeModulesDir, name, exactVersion);
                    registerVersion(name, version || `^${exactVersion}`);
                    continue installDeps;
                }
            }
            exactVersion = npmVersions[0];
            version = version || `^${exactVersion}`;
            newDeps[name] = version;
            hasNewDeps = true;
            registerVersion(name, version);
        }
    }
    let hasUpdatedPackageJson = false;
    if (options.installDeps.length > 0) {
        let deps = dependencies ? Object.assign({}, dependencies) : {};
        installDeps(deps, options.installDeps);
        packageJson.dependencies = sortDeps(deps);
        hasUpdatedPackageJson = true;
    }
    if (options.installDevDeps.length > 0) {
        let deps = devDependencies ? Object.assign({}, devDependencies) : {};
        installDeps(deps, options.installDevDeps);
        packageJson.devDependencies = sortDeps(deps);
        hasUpdatedPackageJson = true;
    }
    if (options.uninstallDeps.length > 0) {
        if (options.verbose) {
            console.log('uninstalling packages:', options.uninstallDeps);
        }
        for (let dep of options.uninstallDeps) {
            let { name } = parseDep(dep);
            uninstallDep(nodeModulesDir, name);
            if (packageJson.dependencies && name in packageJson.dependencies) {
                delete packageJson.dependencies[name];
                hasUpdatedPackageJson = true;
                if (dependencies) {
                    delete dependencies[name];
                }
            }
            if (packageJson.devDependencies && name in packageJson.devDependencies) {
                delete packageJson.devDependencies[name];
                hasUpdatedPackageJson = true;
                if (devDependencies) {
                    delete devDependencies[name];
                }
            }
        }
    }
    if (hasUpdatedPackageJson) {
        (0, fs_1.writeFileSync)(packageFile, JSON.stringify(packageJson, null, 2));
    }
    function addPackageDep(name, versionRange) {
        if (isLinkDep(versionRange)) {
            linkDepPackage(linkedDepPackageDirs, nodeModulesDir, name, versionRange);
            getVersions(usedPackageVersions, name).add(versionRange);
            return;
        }
        let versions = Array.from(getVersions(storePackageVersions, name));
        let exactVersion = findLatestMatch(versionRange, versions);
        if (exactVersion) {
            linkStorePackage(storeDir, nodeModulesDir, name, exactVersion);
            return;
        }
        newDeps[name] = versionRange;
        hasNewDeps = true;
    }
    if (devDependencies && options.dev) {
        for (let name in devDependencies) {
            let version = devDependencies[name];
            addPackageDep(name, version);
        }
    }
    if (dependencies) {
        for (let name in dependencies) {
            let version = dependencies[name];
            addPackageDep(name, version);
        }
    }
    function collectNodeModules(nodeModulesDir) {
        // detect cyclic dependencies
        let realNodeModulesDir = (0, fs_1.realpathSync)(nodeModulesDir);
        if (collectedNodeModules.has(realNodeModulesDir))
            return;
        collectedNodeModules.add(realNodeModulesDir);
        for (let dirname of (0, fs_1.readdirSync)(nodeModulesDir)) {
            if (dirname[0] === '.')
                continue;
            if (dirname[0] !== '@') {
                let packageDir = (0, path_1.join)(nodeModulesDir, dirname);
                collectPackage(packageDir);
                continue;
            }
            let orgName = dirname;
            let orgDir = (0, path_1.join)(nodeModulesDir, orgName);
            for (let dirname of (0, fs_1.readdirSync)(orgDir)) {
                let packageDir = (0, path_1.join)(orgDir, dirname);
                collectPackage(packageDir);
            }
        }
    }
    function collectPackage(packageDir) {
        let stats = (0, fs_1.lstatSync)(packageDir);
        if (stats.isSymbolicLink())
            return;
        let { json: { name, version }, file, } = getPackageJson(packageDir);
        if (!name)
            throw new Error(`missing package name in ${file}`);
        if (!version)
            throw new Error(`missing package version in ${file}`);
        getVersions(usedPackageVersions, name).add(version);
        let realPackageDir = (0, fs_1.realpathSync)(packageDir);
        if (linkedDepPackageDirs.has(realPackageDir))
            return;
        getVersions(storePackageVersions, name).add(version);
        let nodeModulesDir = (0, path_1.join)(packageDir, 'node_modules');
        if ((0, fs_1.existsSync)(nodeModulesDir)) {
            collectNodeModules(nodeModulesDir);
        }
        let key = `${name}@${version}`;
        let storePackageDir = (0, path_1.join)(storeDir, key);
        if ((0, fs_1.existsSync)(storePackageDir)) {
            (0, fs_1.rmSync)(packageDir, { recursive: true });
            return;
        }
        if (name.includes('/')) {
            let parentDir = (0, path_1.dirname)(storePackageDir);
            (0, fs_1.mkdirSync)(parentDir, { recursive: true });
        }
        mv(packageDir, storePackageDir);
    }
    if (hasNewDeps) {
        if (options.verbose) {
            console.log('installing new packages:', newDeps);
        }
        let tmpDir = (0, path_1.join)(nodeModulesDir, '.tmp');
        (0, fs_1.mkdirSync)(tmpDir, { recursive: true });
        npmInstall(context, tmpDir, newDeps);
        let tmpNodeModulesDir = (0, path_1.join)(tmpDir, 'node_modules');
        collectNodeModules(tmpNodeModulesDir);
    }
    collectNodeModules(nodeModulesDir);
    if (options.verbose && usedPackageVersions.size > 0) {
        console.log('linking packages:', usedPackageVersions);
    }
    let linkedDeps = new Set();
    function linkDeps(packageDir, dependencies) {
        // detect cyclic dependencies
        let realPackageDir = (0, fs_1.realpathSync)(packageDir);
        if (linkedDeps.has(realPackageDir))
            return;
        linkedDeps.add(realPackageDir);
        let nodeModulesDir = (0, path_1.join)(packageDir, 'node_modules');
        let hasNodeModulesDir = false;
        for (let name in dependencies) {
            if (!hasNodeModulesDir) {
                (0, fs_1.mkdirSync)(nodeModulesDir, { recursive: true });
                hasNodeModulesDir = true;
            }
            let versionRange = dependencies[name];
            linkDep(nodeModulesDir, name, versionRange);
        }
    }
    // nodeModulesDir -> name -> depPackageDir
    let depPackageDirs = new Map();
    function linkDep(nodeModulesDir, name, versionRange) {
        if (versionRange.startsWith('npm:')) {
            // e.g. package "@isaacs/cliui@8.0.2" has dependency on "string-width-cjs" with version "npm:string-width@^4.2.0"
            let parts = versionRange.slice(4).split('@');
            if (parts.length == 2) {
                name = parts[0];
                versionRange = parts[1];
            }
            else if (parts.length == 3) {
                name = parts[0] + '@' + parts[1];
                versionRange = parts[2];
            }
            else {
                throw new Error(`Invalid versionRange: ${versionRange}, name: ${name}, nodeModulesDir: ${nodeModulesDir}`);
            }
        }
        if (isLinkDep(versionRange)) {
            let depPackageDir = parseLinkDepPackageDir(versionRange);
            getMap2(depPackageDirs, nodeModulesDir).set(name, depPackageDir);
            let bin = getPackageJson(depPackageDir).json.bin;
            linkBin(nodeModulesDir, depPackageDir, name, bin);
            return;
        }
        let versions = Array.from(getVersions(storePackageVersions, name));
        let exactVersion = findLatestMatch(versionRange, versions);
        if (!exactVersion)
            throw new Error(`missing package, name: ${name}, versionRange: ${versionRange}`);
        let depPackageDir = linkStorePackage(storeDir, nodeModulesDir, name, exactVersion);
        getMap2(depPackageDirs, nodeModulesDir).set(name, depPackageDir);
        let { dependencies, bin } = getPackageJson(depPackageDir).json;
        if (dependencies) {
            linkDeps(depPackageDir, dependencies);
        }
        linkBin(nodeModulesDir, depPackageDir, name, bin);
    }
    for (let name in newInstallDeps) {
        let version = newInstallDeps[name];
        linkDep(nodeModulesDir, name, version);
    }
    if (devDependencies && options.dev) {
        for (let name in devDependencies) {
            let version = devDependencies[name];
            linkDep(nodeModulesDir, name, version);
        }
    }
    if (dependencies) {
        for (let name in dependencies) {
            let version = dependencies[name];
            linkDep(nodeModulesDir, name, version);
        }
    }
    let linkedPeerDeps = new Set();
    function linkPeerDeps(nodeModulesDir) {
        try {
            let realNodeModulesDir = (0, fs_1.realpathSync)(nodeModulesDir);
            if (linkedPeerDeps.has(realNodeModulesDir))
                return;
            linkedPeerDeps.add(realNodeModulesDir);
        }
        catch (error) {
            return;
        }
        let parentDeps = getMap2(depPackageDirs, nodeModulesDir);
        parentDeps.forEach(depPackageDir => {
            let { json } = getPackageJson(depPackageDir);
            let nodeModulesDir = (0, path_1.join)(depPackageDir, 'node_modules');
            (0, fs_1.mkdirSync)(nodeModulesDir, { recursive: true });
            let linkedPeerDeps = new Set();
            for (let name in json.peerDependencies) {
                linkPeerDep(parentDeps, nodeModulesDir, name);
                linkedPeerDeps.add(name);
            }
            for (let name in json.peerDependenciesMeta) {
                if (linkedPeerDeps.has(name))
                    return;
                linkPeerDep(parentDeps, nodeModulesDir, name);
            }
            linkPeerDeps(nodeModulesDir);
        });
    }
    linkPeerDeps(nodeModulesDir);
}
function linkPeerDep(parentDeps, nodeModulesDir, name) {
    let peerDepDir = parentDeps.get(name);
    if (!peerDepDir)
        return;
    let depDir = (0, path_1.join)(nodeModulesDir, name);
    if (name.includes('/')) {
        let parentDir = (0, path_1.dirname)(depDir);
        (0, fs_1.mkdirSync)(parentDir, { recursive: true });
    }
    makeSymbolicLink(peerDepDir, depDir);
}
function getVersions(map, name) {
    let set = map.get(name);
    if (!set) {
        set = new Set();
        map.set(name, set);
    }
    return set;
}
function getMap2(map, key) {
    let map2 = map.get(key);
    if (!map2) {
        map2 = new Map();
        map.set(key, map2);
    }
    return map2;
}
// packageDir -> PackageJSON
let jsonCache = new Map();
function getPackageJson(packageDir) {
    let entry = jsonCache.get(packageDir);
    if (entry)
        return entry;
    let file = (0, path_1.join)(packageDir, 'package.json');
    let json = JSON.parse((0, fs_1.readFileSync)(file).toString());
    entry = { json, file };
    jsonCache.set(packageDir, entry);
    return entry;
}
function findLatestMatch(versionRange, exactVersions) {
    if (versionRange === 'latest') {
        versionRange = '*';
    }
    return semver_1.default.maxSatisfying(exactVersions, versionRange);
}
function makeSymbolicLink(src, dest) {
    try {
        (0, fs_1.symlinkSync)(src, dest);
    }
    catch (err) {
        if (err.code === 'EEXIST') {
            return;
        }
        throw err;
    }
}
function linkStorePackage(storeDir, nodeModulesDir, packageName, exactVersion) {
    let src = (0, path_1.join)(storeDir, `${packageName}@${exactVersion}`);
    let dest = (0, path_1.join)(nodeModulesDir, packageName);
    linkPackage(packageName, src, dest);
    return src;
}
function linkDepPackage(linkedDepPackageDirs, nodeModulesDir, packageName, dep) {
    let src = parseLinkDepPackageDir(dep);
    if (!packageName) {
        let { json, file } = getPackageJson(src);
        if (!json.name)
            throw new Error(`missing package name in ${file}`);
        packageName = json.name;
    }
    let dest = (0, path_1.join)(nodeModulesDir, packageName);
    linkPackage(packageName, src, dest);
    let packageDir = (0, fs_1.realpathSync)(dest);
    linkedDepPackageDirs.add(packageDir);
    return packageName;
}
function linkPackage(packageName, src, dest) {
    if (packageName.includes('/')) {
        let parentDir = (0, path_1.dirname)(dest);
        (0, fs_1.mkdirSync)(parentDir, { recursive: true });
    }
    makeSymbolicLink(src, dest);
}
function linkBin(nodeModulesDir, packageDir, name, bin) {
    if (!bin)
        return;
    let binDir = (0, path_1.join)(nodeModulesDir, '.bin');
    (0, fs_1.mkdirSync)(binDir, { recursive: true });
    if (typeof bin === 'string') {
        if (name[0] === '@') {
            name = name.split('/')[1];
        }
        linkBinFile(binDir, packageDir, name, bin);
        return;
    }
    for (let name in bin) {
        let filename = bin[name];
        linkBinFile(binDir, packageDir, name, filename);
    }
}
function linkBinFile(binDir, packageDir, name, filename) {
    let src = (0, path_1.join)(packageDir, filename);
    let dest = (0, path_1.join)(binDir, name);
    setBinPermission(src);
    makeSymbolicLink(src, dest);
}
let setBinPermissionCache = new Set();
let binPrefix = Buffer.from('#!/usr/bin/env node' + os_1.EOL);
let binPrefixCode = binPrefix[0];
let binBuffer = Buffer.alloc(1);
let binFlag = fs_2.default.constants.X_OK;
function setBinPermission(file) {
    if (setBinPermissionCache.has(file))
        return;
    setBinPermissionCache.add(file);
    try {
        (0, fs_1.accessSync)(file, binFlag);
        return;
    }
    catch (error) {
        /* read first byte */
        binBuffer[0] = 0;
        let fd = (0, fs_1.openSync)(file, 'r');
        (0, fs_1.readSync)(fd, binBuffer);
        (0, fs_1.closeSync)(fd);
        /* insert executable line if not already present */
        if (binBuffer[0] !== binPrefixCode) {
            let content = (0, fs_1.readFileSync)(file);
            let tmpFile = file + '.tmp';
            fd = (0, fs_1.openSync)(tmpFile, 'w+');
            (0, fs_1.writeSync)(fd, binPrefix);
            (0, fs_1.writeSync)(fd, content);
            (0, fs_1.closeSync)(fd);
            mv(tmpFile, file);
        }
        /* set executable permission */
        (0, fs_1.chmodSync)(file, 0o755);
    }
}
function npmInstall(context, cwd, dependencies) {
    let cmd = 'npx npm i';
    if (context.options.legacyPeerDeps) {
        cmd += ' --legacy-peer-deps';
    }
    let file = (0, path_1.join)(cwd, 'package.json');
    let json = { dependencies };
    let text = JSON.stringify(json);
    (0, fs_1.writeFileSync)(file, text);
    (0, child_process_1.execSync)(cmd, { cwd });
}
function mv(src, dest) {
    let cmd = `mv ${JSON.stringify(src)} ${JSON.stringify(dest)}`;
    (0, child_process_1.execSync)(cmd);
}
function parseDep(dep) {
    if (dep.length === 0) {
        throw new Error('Invalid dependency format (empty string)');
    }
    let parts = dep.split('@');
    switch (parts.length) {
        case 1:
            // e.g. semver
            return { name: parts[0], version: null };
        case 2:
            if (parts[0].length === 0) {
                // e.g. @types/semver
                return { name: '@' + parts[1], version: null };
            }
            // e.g. semver@^7.3.7
            return {
                name: parts[0],
                version: parts[1] || null,
            };
        case 3:
            if (parts[0].length > 0)
                throw new Error('Invalid dependency format: ' + JSON.stringify(dep));
            // e.g. @types/semver@^7.3.9
            return {
                name: '@' + parts[1],
                version: parts[2] || null,
            };
        default:
            throw new Error('Invalid dependency format: ' + JSON.stringify(dep));
    }
}
function npmViewVersions(dep) {
    let cmd = `npm view ${JSON.stringify(dep)} version`;
    let stdout = (0, child_process_1.execSync)(cmd);
    let versions = [];
    stdout
        .toString()
        .split('\n')
        .forEach(line => {
        // e.g. `semver@1.0.8 '1.0.8'`
        let version = line.trim().split(' ').pop();
        if (!version)
            return;
        versions.push(version.replace(/'/g, ''));
    });
    return versions;
}
function uninstallDep(nodeModulesDir, name) {
    let dir = (0, path_1.join)(nodeModulesDir, name);
    (0, fs_1.rmSync)(dir, { recursive: true, force: true });
}
function sortDeps(deps) {
    let newDeps = {};
    Object.keys(deps)
        .sort()
        .forEach(name => {
        let version = deps[name];
        newDeps[name] = version;
    });
    return newDeps;
}
function scanPackageRecursively(context, dir, visited) {
    let realDir = (0, fs_1.realpathSync)(dir);
    if (visited.has(realDir))
        return;
    visited.add(realDir);
    for (let filename of (0, fs_1.readdirSync)(dir)) {
        if (filename[0] === '.' || filename === 'node_modules')
            continue;
        let file = (0, path_1.join)(dir, filename);
        let stat = (0, fs_1.statSync)(file);
        if (filename === 'package.json' && stat.isFile()) {
            if (context.options.verbose) {
                console.log('installing packages in:', dir);
            }
            installPackages(context, dir);
            continue;
        }
        if (stat.isDirectory()) {
            scanPackageRecursively(context, file, visited);
            continue;
        }
    }
}
function isLinkDep(dep) {
    return dep.startsWith('link:') || dep.startsWith('file:');
}
function parseLinkDepPackageDir(dep) {
    return dep.slice('link:'.length);
}
